# ......................................................................................
# MIT License

# Copyright (c) 2020-2022 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................

from sklearn.model_selection import train_test_split  # import a standalone procedure toyfunction from the pacakge
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from radnn import mlsys
from radnn.data.sample_set_simple import SampleSet
from .constants import DataPreprocessingKind
from enum import Enum


# =========================================================================================================================
class LegacyDataSet(object):
  # --------------------------------------------------------------------------------------
  # Constructor
  def __init__(self, sample_count=None, feature_count=None, class_count=None, random_seed=None):
    # ................................................................
    # // Fields \\
    self.feature_count = feature_count
    self.class_count = class_count
    self.sample_count = sample_count
    self.random_seed = random_seed
    
    self.samples = None
    self.labels = None
    self.preprocessor = None
    self.mean = None
    self.std = None

    # training set object
    self.ts: SampleSet | None = None
    # validation set object
    self.vs: SampleSet | None = None
    # unknown test set object
    self.us: SampleSet | None = None
    # ................................................................
    if self.random_seed is not None:
      mlsys.random_seed_all(self.random_seed)
  # --------------------------------------------------------------------------------------
  def split(self, p_nValidationSamplesPC=0.10, preprocessing: DataPreprocessingKind | None = DataPreprocessingKind.STANDARDIZE):
    oTS_Samples, oVS_Samples, oTS_Labels, oVS_Labels = train_test_split(self.samples, self.labels ,
                                                                        test_size=p_nValidationSamplesPC,
                                                                        random_state=2021)
    if preprocessing == DataPreprocessingKind.MIN_MAX_NORMALIZE:
      self.preprocessor = MinMaxScaler().fit(oTS_Samples)
    elif preprocessing == DataPreprocessingKind.STANDARDIZE:
      self.preprocessor = StandardScaler().fit(oTS_Samples)

    if self.preprocessor is not None:
      oTS_Samples = self.preprocessor.transform(oTS_Samples)
      oVS_Samples = self.preprocessor.transform(oVS_Samples)

    # (Re)creating the subsets of the dataset after the splits have been created
    self.ts = SampleSet(self, oTS_Samples, oTS_Labels)
    self.vs = SampleSet(self, oVS_Samples, oVS_Labels)
    
    print("%d samples in the Training Set" % self.ts.sample_count)
    print("%d samples in the Validation Set" % self.vs.sample_count)
  # --------------------------------------------------------------------------------------
  # Backwards Compatibility
  @property
  def TSSamples(self):
    return self.ts.samples

  @property
  def ts_samples(self):
    return self.ts_samples

  @property
  def TSLabels(self):
    return self.ts.labels

  @property
  def ts_labels(self):
    return self.ts.labels

  @property
  def TSSampleCount(self):
    return self.ts.sample_count

  @property
  def ts_sample_count(self):
    return self.ts.sample_count

  def VSSamples(self):
    return self.vs.samples

  @property
  def vs_samples(self):
    return self.vs.samples

  @property
  def VSLabels(self):
    return self.vs.labels

  @property
  def vs_labels(self):
    return self.vs.labels

  @property
  def VSSampleCount(self):
    return self.vs.sample_count

  @property
  def vs_sample_count(self):
    return self.vs.sample_count

  @property
  def FeatureCount(self):
    return self.feature_count

  @property
  def ClassCount(self):
    return self.class_count

  @property
  def ClassCount(self):
    return self.class_count

  @property
  def SampleCount(self):
    return self.sample_count

  @property
  def Samples(self):
    return self.samples

  @property
  def Labels(self):
    return self.labels
# =========================================================================================================================



from PIL import Image
from torchvision.transforms import functional as trf



# --------------------------------------------------------------------------------------------------------------------
def resize_largest_side(img: Image.Image, target_size):
  w, h = img.size
  if w > h:  # width is larger
    new_w = target_size
    new_h = int(h * target_size / w)
  else:  # height is larger or equal
    new_h = target_size
    new_w = int(w * target_size / h)
  return trf.resize(img, [new_h, new_w], interpolation=trf.InterpolationMode.BICUBIC)  # use functional resize
# --------------------------------------------------------------------------------------------------------------------
def pad_to_square(img, target_size, fill=0):
  w, h = img.size
  pad_left = (target_size - w) // 2
  pad_top = (target_size - h) // 2
  pad_right = target_size - w - pad_left
  pad_bottom = target_size - h - pad_top
  return trf.pad(img, [pad_left, pad_top, pad_right, pad_bottom], fill=fill)
# --------------------------------------------------------------------------------------------------------------------


# ......................................................................................
# MIT License

# Copyright (c) 2019-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
import numpy as np
from radnn import FileStore
from .dataset_base import DataSetBase

class ImageDataSet(DataSetBase):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, fs, name=None, variant=None, image_shape=None, random_seed=None, is_classification=True):
    super(ImageDataSet, self).__init__(fs, name, variant, random_seed, is_classification)

    self.image_shape = None
    self.feature_count = None
    if image_shape is not None:
      self.image_shape = image_shape
      self.feature_count = int(np.prod(list(image_shape)))

    self.source_fs = None
    self.source_class_subfs_list = None
    self.image_file_list = None
  # --------------------------------------------------------------------------------------------------------------------
  def build(self, source_fs):
    self.source_fs = source_fs
    # If a path is supplied init the file store
    if isinstance(source_fs, str):
      self.source_fs = FileStore(source_fs)

    if self.is_classification:
      self._determine_classes()
      self._detect_class_subfilestores()

    self._generate_image_file_list()
    #TODO: Image preprocess / resize / save shard

    return self
  # --------------------------------------------------------------------------------------------------------------------
  def _determine_classes(self):

    if self.class_names is None:
      # TODO: Detect a JSON file that has the class names dictionary on the source filestore
      pass
      # TODO: Enumerate all subfolder in the source filestore

  # --------------------------------------------------------------------------------------------------------------------
  def _detect_class_subfilestores(self):
    self.source_class_subfs_list = []
  # --------------------------------------------------------------------------------------------------------------------
  def _generate_image_file_list(self):
    self.image_file_list = []
  # --------------------------------------------------------------------------------------------------------------------
  def preview_images(self):
    import matplotlib.pyplot as plt

    # Look at some sample images from dataset
    plt.figure(figsize=(10, 10))
    for i in range(25):
      plt.subplot(5, 5, i + 1)
      if self.ts_samples.shape[2] == 2:
        plt.imshow(self.ts_samples[i].squeeze().astype(np.uint8), cmap='gray')
      else:
        plt.imshow(self.ts_samples[i].squeeze().astype(np.uint8))

      nClassIndex = self.ts_labels[i]
      sClassDescr = str(nClassIndex)
      if isinstance(self.class_names, dict):
        if nClassIndex in self.class_names:
          sClassName = self.class_names[nClassIndex]
          sClassDescr += " `" + sClassName + "`"
      elif isinstance(self.class_names, list):
        sClassName = self.class_names[nClassIndex]
        sClassDescr += " `" + sClassName + "` "

      plt.title(f"Label: {sClassDescr}")
      plt.axis('off')
    plt.show()
  # --------------------------------------------------------------------------------------------------------------------



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import shutil
import glob
import sys
import zipfile
from typing import Any, Type
import importlib.util
if (sys.version_info.major == 3) and (sys.version_info.minor <= 7):
  import pickle5 as pickle
else:
  import pickle
from radnn.system.files import FileList
from radnn.system.files import JSONFile
from radnn.system.files import PickleFile
from radnn.system.files import TextFile
from radnn.system.files import CSVFile
from radnn.system.files import ZipFile
from radnn.errors import *


_is_opencv_installed = importlib.util.find_spec("cv2") is not None
if _is_opencv_installed:
  from radnn.system.files.imgfile import PNGFile

# =======================================================================================================================
class FileStore(object):
  # --------------------------------------------------------------------------------------------------------
  def __init__(self, base_folder, is_verbose=False, must_exist=False):
    #.......................... |  Instance Attributes | ............................
    self.base_folder = base_folder
    self.absolute_path = os.path.abspath(base_folder)
    if not os.path.exists(self.base_folder):
      if must_exist:
        raise Exception(FS_MUST_EXIST % self.base_folder)
      else:
        os.makedirs(self.base_folder)

    self.is_verbose = is_verbose
    self.json: JSONFile = JSONFile(None, parent_folder=self.base_folder)
    self.obj: PickleFile = PickleFile(None, parent_folder=self.base_folder)
    self.text: TextFile = TextFile(None, parent_folder=self.base_folder)
    self.csv: CSVFile = CSVFile(None, parent_folder=self.base_folder)
    if _is_opencv_installed:
      self.img: PNGFile = PNGFile(None, parent_folder=base_folder)
    self.donefs = None
    #................................................................................
  # --------------------------------------------------------------------------------------------------------
  @property
  def has_files(self):
    return not self.is_empty
  # --------------------------------------------------------------------------------------------------------
  @property
  def is_empty(self):
    bExists = not os.path.exists(self.base_folder)
    if not bExists:
      bExists = len(os.listdir(self.base_folder)) > 0
    return not bExists
  # --------------------------------------------------------------------------------------------------------
  def exists_folder(self, filename):
    sFullPath = os.path.join(self.base_folder, filename)
    return os.path.exists(sFullPath)
  # --------------------------------------------------------------------------------------------------------
  def exists(self, filename):
    sFullFilePath = os.path.join(self.base_folder, filename)
    return os.path.isfile(sFullFilePath)
  # --------------------------------------------------------------------------------------------------------
  def subfs(self, subfolder_name, must_exist=False):
    return FileStore(self.subpath(subfolder_name), must_exist=must_exist)
  # --------------------------------------------------------------------------------------------------------
  def subpath(self, subfolder_name):
    if os.path.sep == "\\":
      if subfolder_name.find(os.path.sep) < 0:
        subfolder_name = subfolder_name.replace("/", "\\")
    return self.folder(subfolder_name)
  # --------------------------------------------------------------------------------------------------------
  def folder(self, folder_name):
    sFolder = os.path.join(self.absolute_path, folder_name)
    if not os.path.exists(sFolder):
      os.makedirs(sFolder)

    return sFolder
  # --------------------------------------------------------------------------------------------------------
  def file(self, file_name, file_ext=None):
    if file_ext is not None:
      if file_ext.find(".") < 0:
        file_name += "." + file_ext
      else:
        file_name += file_ext
    return os.path.join(self.absolute_path, file_name)
  # --------------------------------------------------------------------------------------------------------
  def get_file_kind(self, file_name: str):
    name, ext = os.path.splitext(file_name)
    ext = ext.lower()
    if ext == ".pkl":
      return "obj"
    elif ext == ".txt":
      return "text"
    elif ext == ".json":
      return "json"
    elif ext == ".png":
      return "img"
    elif ext == ".zip":
      return "zip"
    else:
      return "?"
  # --------------------------------------------------------------------------------------------------------
  def artifact(self, file_name: str, kind: str | None = None) -> Any:
    if kind is None:
      kind = self.get_file_kind(file_name)
      
    oFile = None
    if kind == "obj":
      oFile = PickleFile(file_name, parent_folder=self.base_folder)
    elif kind == "text":
      oFile = TextFile(file_name, parent_folder=self.base_folder)
    elif kind == "json":
      oFile = JSONFile(file_name, parent_folder=self.base_folder)
    elif kind == "csv":
      oFile = CSVFile(file_name, parent_folder=self.base_folder)
    elif kind == "img":
      if _is_opencv_installed:
        oFile = PNGFile(file_name, parent_folder=self.base_folder)
      else:
        print("Open CV is not installed")  #TODO: Support PIL
    elif kind == "zip":
      oFile = ZipFile(file_name, parent_folder=self.base_folder)
      
    if oFile is None:
      raise Exception(FILESTORE_DATAFILE_KIND_NOT_SUPPORTED % str(kind))
    
    return oFile
  # --------------------------------------------------------------------------------------------------------
  def entries(self):
    return os.listdir(self.base_folder)
  # --------------------------------------------------------------------------------------------------------
  def _ls(self, file_matching_pattern, is_removing_extension, sort_filename_key):
    if ";" in file_matching_pattern:
      sEntries = []
      for sExt in file_matching_pattern.split(";"):
        sEntries += glob.glob1(self.base_folder, sExt.strip())
    else:
      sEntries = glob.glob1(self.base_folder, file_matching_pattern)

    if is_removing_extension:
      oFileNamesOnly = []
      for sEntry in sEntries:
        sFileNameOnly, _ = os.path.splitext(sEntry)
        oFileNamesOnly.append(sFileNameOnly)
      sEntries = sorted(oFileNamesOnly, key=sort_filename_key)

    return sEntries
  # --------------------------------------------------------------------------------------------------------
  def list_files(self, file_matching_pattern, is_full_path=True, is_removing_extension=False, sort_filename_key=None):
    sEntries = self._ls(file_matching_pattern, is_removing_extension, sort_filename_key)
    if is_full_path:
      oResult = [os.path.join(self.base_folder, x) for x in sEntries]
    else:
      oResult = [x for x in sEntries]
    return oResult
  # --------------------------------------------------------------------------------------------------------
  def filelist(self, file_matching_pattern, is_removing_extension=False, sort_filename_key=None):
    sEntries = self._ls(file_matching_pattern, is_removing_extension, sort_filename_key)
    oFileList = FileList(self.base_folder)
    for x in sEntries:
      oFileList.append(x)
    return oFileList
  # --------------------------------------------------------------------------------------------------------
  def list_folders(self, is_full_path=True):
    sResult = []
    for sFolder in os.listdir(self.base_folder):
      sFullPath = os.path.join(self.base_folder, sFolder)
      if os.path.isdir(sFullPath):
        if is_full_path:
          sResult.append(sFullPath)
        else:
          sResult.append(sFolder)
    return sResult
  # --------------------------------------------------------------------------------------------------------
  def dequeue_file(self, file_matching_pattern, is_full_path=True, archive_folder_name=".done"):
    if self.donefs is None:
      self.donefs = self.subfs(archive_folder_name)
    oQueue = self.files(file_matching_pattern, is_full_path=False)

    if len(oQueue) == 0:
      sFileName = None
    else:
      sFileName = oQueue[0]
      shutil.move(os.path.join(self.base_folder, sFileName), os.path.join(self.donefs.base_folder, sFileName))

    if is_full_path and (sFileName is not None):
      sFileName = os.path.join(self.donefs.base_folder, sFileName)
    return sFileName
  
  # --------------------------------------------------------------------------------------------------------
  def move_to(self, source_files: str | list, fs):
    oFiles = source_files
    if isinstance(source_files, str):
      # A pattern is given
      if "*" in source_files:
        oFiles = self.list_files(source_files, is_full_path=False)
      else:
        oFiles = [source_files]
    oDestFiles = []
    for sFile in oFiles:
      sDestFile =  os.path.join(fs.absolute_path, sFile)
      oDestFiles.append(sDestFile)
      shutil.move(os.path.join(self.absolute_path, sFile),sDestFile)
    return oDestFiles
  
  # --------------------------------------------------------------------------------------------------------
  def copy_to(self, source_files: str | list, fs):
    oFiles = source_files
    if isinstance(source_files, str):
      # A pattern is given
      if "*" in source_files:
        oFiles = self.list_files(source_files, is_full_path=False)
      else:
        oFiles = [source_files]
    oDestFiles = []
    for sFile in oFiles:
      sDestFile =  os.path.join(fs.absolute_path, sFile)
      oDestFiles.append(sDestFile)
      shutil.copy(os.path.join(self.absolute_path, sFile),sDestFile)
    return oDestFiles
  
  # --------------------------------------------------------------------------------------------------------
  def remove_existing(self, filename):
    sFileName = self.file(filename)
    if os.path.exists(sFileName):
      os.remove(sFileName)
  # --------------------------------------------------------------------------------------------------------
  def compress_to_zip(self, output_zip_path, must_replace=True):
    """
    Compress a folder (including subfolders) into a .zip file.

    Args:
        output_zip_path (str): The path where the .zip file will be saved.
    """
    if must_replace:
      if os.path.exists(output_zip_path):
        os.remove(output_zip_path)
    
      # Create the zip file
      with zipfile.ZipFile(output_zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(self.absolute_path):
          for file in files:
            abs_path = os.path.join(root, file)
            # Get relative path for proper folder structure inside zip
            rel_path = os.path.relpath(abs_path, start=self.absolute_path)
            zipf.write(abs_path, rel_path)
  # --------------------------------------------------------------------------------------------------------
  def extract_from_zip(self, zip_filename, has_progress_bar=True):
    """
    Extracts a .zip file into the specified folder.

    Args:
        zip_filename (str): Path to the .zip file.
        destination_folder (str): Folder where the contents should be extracted.
    """
    # Ensure the zip file exists
    if not os.path.isfile(zip_filename):
      raise FileNotFoundError(f"Zip file not found: {zip_filename}")
    
    sDestFolder = self.absolute_path
    # Ensure destination directory exists
    os.makedirs(sDestFolder, exist_ok=True)
    
    # Extract the zip file
    with zipfile.ZipFile(zip_filename, 'r') as zip_ref:
      file_list = zip_ref.infolist()
      total_files = len(file_list)
      
      oIterator = file_list
      if has_progress_bar:
        from tqdm import tqdm
        oIterator = tqdm(file_list, total=total_files, desc="Extracting", unit="file")
      
      for file_info in oIterator:
        zip_ref.extract(file_info, sDestFolder)
    
    print(f" |_ Extracted '{zip_filename}' to '{sDestFolder}'")
  # --------------------------------------------------------------------------------------------------------
  def purge_done(self):
    #//TODO: Remove from the current filestore all files that are moved into the .done filestore
    pass
  # ----------------------------------------------------------------------------------
  def __repr__(self)->str:
    return self.absolute_path
  # --------------------------------------------------------------------------------------------------------
  def __str__(self)->str:
    return self.absolute_path
  # --------------------------------------------------------------------------------------------------------
# ======================================================================================================================








from radnn import FileStore
from radnn.utils import camel_case
from radnn.system.files import FileList
from .dataset_folder import DataSetFolder
import sys
from tqdm import tqdm
from datetime import datetime

class ImageDataSetFiles(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, images_fs, name="files", is_progress_indicator=True):
    self.images_fs :FileStore = images_fs
    self.name = name
    self.is_progress_indicator = is_progress_indicator
    self.class_names :dict = dict()
    self.class_folders :list = []
    self.files :FileList = dict()
    self.files_ts :FileList = dict()
    self.files_vs :FileList = dict()
    self.files_ut :FileList = dict()
    self.total_file_count = 0
    self.is_split_on_main_folder = False
    self.is_split_in_class_folders = False
    self.run_date_time = None

    self.detect_class_names_from_folders()
  # --------------------------------------------------------------------------------------------------------------------
  def load(self, extensions="*.jpg;  *.png"):
    if not self.load_file_lists():
      self.detect_files(extensions)
  # --------------------------------------------------------------------------------------------------------------------
  def load_file_lists(self):
    bResult = False
    oDict = self.images_fs.obj.load(f"{self.name}-file-info.pkl")
    if oDict is not None:
      self.run_date_time = oDict["RunDateTime"]
      self.class_names = oDict["ClassNames"]
      self.class_folders = oDict["ClassFolders"]
      self.files = oDict["ClassFiles.All"]
      self.files_ts = oDict["ClassFiles.TrainingSet"]
      self.files_vs = oDict["ClassFiles.ValidationSet"]
      self.files_ut = oDict["ClassFiles.UnknownTestSet"]
      self.total_file_count = oDict["TotalFileCount"]
      self.is_split_on_main_folder = oDict["IsSplitOnMainFolder"]
      self.is_split_in_class_folders = oDict["IsSplitInClassFolders"]
      bResult = True

    return bResult
  # --------------------------------------------------------------------------------------------------------------------
  def save_file_lists(self):
    oDict = dict()
    oDict["RunDateTime"] = self.run_date_time
    oDict["ClassNames"] = self.class_names
    oDict["ClassFolders"] = self.class_folders
    oDict["ClassFiles.All"] = self.files
    oDict["ClassFiles.TrainingSet"] = self.files_ts
    oDict["ClassFiles.ValidationSet"] = self.files_vs
    oDict["ClassFiles.UnknownTestSet"] = self.files_ut
    oDict["TotalFileCount"] = self.total_file_count
    oDict["IsSplitOnMainFolder"] = self.is_split_on_main_folder
    oDict["IsSplitInClassFolders"] = self.is_split_in_class_folders
    self.images_fs.obj.save(oDict, f"{self.name}-file-info.pkl")
  # --------------------------------------------------------------------------------------------------------------------
  def detect_class_names_from_folders(self):
    oClassNamesFS = self.images_fs
    oMainFolder = DataSetFolder("/", self.images_fs)
    oFolders = oMainFolder.subfolders

    self.is_split_on_main_folder = oMainFolder.is_split
    if self.is_split_on_main_folder:
      # Detect the class names under the training set subfolder
      oClassNamesFS = oMainFolder.filestore_ts
      oFolders = oClassNamesFS.list_folders(is_full_path=False)

    for nIndex, sFolder in enumerate(oFolders):
      sClassName = camel_case(sFolder)
      self.class_names[nIndex] = sClassName
      oClassFS = oClassNamesFS.subfs(sFolder, must_exist=True)
      oClassFolder = DataSetFolder(sFolder, oClassFS)
      if not self.is_split_on_main_folder:
        if oClassFolder.is_split:
          self.is_split_in_class_folders = True
      self.class_folders.append(oClassFolder)

    return self.class_folders

  # --------------------------------------------------------------------------------------------------------------------
  def traverse_sub_folders(self, extensions, progress):
    for nClassIndex, oClassFolder in enumerate(self.class_folders):
      if progress is not None:
        progress.set_description(f"Finding files for class {self.class_names[nClassIndex]}")
        progress.refresh()
      self.files[nClassIndex] = oClassFolder.filestore.filelist(extensions)
      self.total_file_count += len(self.files[nClassIndex])
      if progress is not None:
        progress.update(1)
  # --------------------------------------------------------------------------------------------------------------------
  def traverse_sub_folders_with_split(self, extensions, progress):
    self.total_file_count = 0
    for nClassIndex, oClassFolder in enumerate(self.class_folders):
      if progress is not None:
        progress.set_description(f"Finding files for class {self.class_names[nClassIndex]}")
        progress.refresh()
      if oClassFolder.is_split:
        oClassAllFiles = FileList()
        for nIndex, oSplitFileStore in enumerate(oClassFolder.split_filestores):
          if oSplitFileStore is not None:
            oFileList = oSplitFileStore.filelist(extensions)
            for oFile in oFileList.full_paths:
              dSplit = None
              if oSplitFileStore == oClassFolder.filestore_ts:
                dSplit = self.files_ts
              elif oSplitFileStore == oClassFolder.filestore_vs:
                dSplit = self.files_vs
              elif oSplitFileStore == oClassFolder.filestore_ut:
                dSplit = self.files_ut

              if dSplit is not None:
                if nClassIndex not in dSplit:
                  dSplit[nClassIndex] = []
                dSplit[nClassIndex].append(oFile)

              oClassAllFiles.append(oFile)
      else:
        raise Exception(f"No split subfolders for class {nIndex} '{self.class_names[nIndex]}',\n"
                        + f"that is stored in {oClassFolder.filestore}\n"
                        + f"All of the classes should have the same split.")
      self.files[nClassIndex] = oClassAllFiles
      self.total_file_count += len(self.files[nClassIndex])
      if progress is not None:
        progress.update(1)


    if progress is not None:
      progress.set_description("Finished")
      progress.refresh()
  # --------------------------------------------------------------------------------------------------------------------
  def detect_files(self, extensions="*.jpg;  *.png"):
    oProgress = None
    if len(self.class_folders) > 0:
      if (not self.is_split_on_main_folder) and (not self.is_split_in_class_folders):
        if self.is_progress_indicator:
          oProgress = tqdm(total=len(self.class_folders), ncols=80)
        try:
          self.traverse_sub_folders(extensions, oProgress)
        finally:
          if self.is_progress_indicator:
            oProgress.close()

      elif (not self.is_split_on_main_folder) and self.is_split_in_class_folders:
        if self.is_progress_indicator:
          oProgress = tqdm(total=len(self.class_folders), ncols=80)
        try:
          self.traverse_sub_folders_with_split(extensions, oProgress)
        finally:
          if self.is_progress_indicator:
            oProgress.close()

      self.save_file_lists()
  # --------------------------------------------------------------------------------------------------------------------


















# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2020-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import matplotlib.pyplot as plt

class PlotLearningCurve(object):
  # --------------------------------------------------------------------------------------
  def __init__(self, metrics_dict, model_name):
    self.metrics_dict = metrics_dict
    self.model_name = model_name
    print("Keys in training process log:", self.metrics_dict.keys())
  # --------------------------------------------------------------------------------------
  def prepare(self, metric_key="accuracy", custom_title=None, metric_low=0, metric_high=1.0, is_legend_right=False, is_keras=False):
    sTrainMetricKey = metric_key
    if not is_keras:
      sTrainMetricKey = "train_" + metric_key
    sValMetricKey = "val_" + metric_key
    oLegend = ["training"]
    bHasValidation = sValMetricKey in self.metrics_dict
    oLegend.append("validation")
    
    
    plt.clf()
    plt.plot(self.metrics_dict[sTrainMetricKey])
    if bHasValidation:
      plt.plot(self.metrics_dict[sValMetricKey])
      
    if custom_title is None:
      plt.title(self.model_name + ' ' + metric_key)
    else:
      plt.title(self.model_name + ' ' + custom_title)
      
    plt.ylabel(metric_key.capitalize().title())
    plt.xlabel("Epoch")
    if (metric_low is not None) and (metric_high is not None):
      plt.ylim([metric_low, metric_high])
    if is_legend_right:
      plt.legend(oLegend, loc="upper right")
    else:
      plt.legend(oLegend, loc="upper left")
    return self
  # --------------------------------------------------------------------------------------
  def prepare_cost(self, cost_function=None):
    if isinstance(cost_function, str):
      sCostFunctionName = cost_function
    else:
      sClassName = str(cost_function.__class__)
      if ("keras" in sClassName) and ("losses" in sClassName):
        sCostFunctionNameParts = cost_function.name.split("_")
        sCostFunctionNameParts = [x.capitalize() + " " for x in sCostFunctionNameParts]
        sCostFunctionName = " ".join(sCostFunctionNameParts)

    return self.prepare("loss", sCostFunctionName, True)
  # --------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------





from radnn.core import MLInfrastructure

if MLInfrastructure.is_windows():
  from .windows_host import WindowsHost

if MLInfrastructure.is_colab():
  from .colab_host import ColabHost

if MLInfrastructure.is_linux():
  from .linux_host import LinuxHost


# Version 0.0.3    [2025-01-25]
# Version 0.0.5    [2025-01-26]
# Version 0.0.6    [2025-02-04]
# Version 0.0.7.2  [2025-02-17]
# Version 0.0.7.3  [2025-02-21]
# Version 0.0.8    [2025-02-25]
# Version 0.0.9    [2025-04-15]
# Version 0.1.0    [2026-01-07]
# Version 0.1.1    [2025-01-08]
# Version 0.1.4    [2025-01-26]
# Version 0.1.5    [2025-02-02]
# Version 0.1.6    [2025-02-03]
# Version 0.1.7    [2025-02-04]
__version__ = "0.1.7"

from .system import FileStore, FileSystem
from .ml_system import MLSystem
from .ml_system import mlsys
from .utils import print_tensor, order_str
from .errors import Errors
from .learn.constants import MLTask



# ......................................................................................
# MIT License

# Copyright (c) 2023-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
import numpy as np

'''
    Standardization for rank 3 and above tensors using numpy
'''
class Standardizer(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, name=None, filestore=None):
    # ................................................................
    # // Fields \\
    self.mean = None
    self.std = None
    self.name = name
    self.filestore = filestore
    # ................................................................
    self.load()
  # --------------------------------------------------------------------------------------------------------------------
  def load(self):
    if (self.name is not None) and (self.filestore is not None):
      dStats = self.filestore.obj.load("%s-meanstd.pkl" % self.name)
      if dStats is not None:
        self.mean = dStats["mean"]
        self.std = dStats["std"]
  # --------------------------------------------------------------------------------------------------------------------
  def save(self):
    if (self.name is not None) and (self.filestore is not None):
      dStats = {"mean": self.mean, "std": self.std}
      self.filestore.obj.save(dStats, "%s-meanstd.pkl" % self.name, is_overwriting=True)
  # --------------------------------------------------------------------------------------------------------------------
  def fit(self, data, axis_for_stats=None, is_recalculating=False, is_verbose=False):
    bIsCached = False
    if (self.name is not None) and (self.filestore is not None):
      if self.mean is not None:
        bIsCached = True

    if (not bIsCached) or is_recalculating:
      # Collect statistics with maximum precision
      data = data.astype(np.float64)
      nAxes = list(range(len(data.shape)))
      if axis_for_stats is None:
        nAxes = tuple(nAxes)
      else:
        if axis_for_stats == -1:
          axis_for_stats = nAxes[-1]

        nAxes.remove(axis_for_stats)
        if len(nAxes) == 1:
          nAxes = nAxes[0]
        else:
          nAxes = tuple(nAxes)

      self.mean = np.mean(data, axis=nAxes)
      self.std = np.std(data, axis=nAxes)
      if is_verbose:
        print("  Standardization: mean/std shape:%s" % str(self.mean.shape))
      self.save()

    return self
  # --------------------------------------------------------------------------------------------------------------------
  def fit_transform(self, data, axis_for_stats=-1, is_recalculating=False, is_verbose=False):
    self.fit(data, axis_for_stats, is_recalculating, is_verbose)
    return self.transform(data)
  # --------------------------------------------------------------------------------------------------------------------
  def standardize(self, data):
    return (data - self.mean) / self.std
  # --------------------------------------------------------------------------------------------------------------------
  def destandardize(self, data):
    return (data * self.std) + self.mean
  # --------------------------------------------------------------------------------------------------------------------
  def transform(self, data):
    nStandardizedData = (data - self.mean) / self.std
    return nStandardizedData.astype(data.dtype)
  # --------------------------------------------------------------------------------------------------------------------
  def inverse_transform(self, data):
    nNonStandardizedData = (data * self.std) + self.mean
    return nNonStandardizedData.astype(data.dtype)
  # --------------------------------------------------------------------------------------------------------------------
  def __str__(self):
    return f"Standardizer: mean={self.mean} std={self.std}"
  # --------------------------------------------------------------------------------------------------------------------
  def __repr__(self):
    return self.__str__()
  # --------------------------------------------------------------------------------------------------------------------


import numpy as np
import time
import functools


class LatencyBenchmark(dict):
  def __init__(self):
    self.enabled = False
  
  def stats(self, key):
    nSeries = np.asarray(self.get(key, [0.0]), np.float32)
    nMean = np.mean(nSeries)
    nStd = np.std(nSeries)
    nMax = np.max(nSeries)
    nMin = np.min(nSeries)
    return nMin, nMax, nMean, nStd

    

mlbench = LatencyBenchmark()


def timed_method(name: str | None = None):
  """
  Decorator to measure elapsed time using a high-resolution timer.

  Usage:
      @timeit()
      def f(...):

      @timeit("custom_name")
      def g(...):
  """
  
  def decorator(func):
    label = name or func.__qualname__
    
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      if not mlbench.enabled:
        return func(*args, **kwargs)
      else:
        start = time.perf_counter()
        try:
          return func(*args, **kwargs)
        finally:
          end = time.perf_counter()
          elapsed = end - start
          if not label in mlbench:
            mlbench[label] = []
          mlbench[label].append(elapsed * 1e3)
    
    return wrapper
  
  return decorator


# ......................................................................................
# MIT License

# Copyright (c) 2019-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
from .dataset_base import DataSetBase
from .subset_type import SubsetType
from radnn.data_beta.preprocess import Normalizer, Standardizer

class DataFeed(object):
  def __init__(self, dataset: DataSetBase, subset_type):
    self.subset_type: SubsetType = None
    if isinstance(subset_type, SubsetType):
      self.subset_type = subset_type
    elif isinstance(subset_type, str):
      self.subset_type = SubsetType(subset_type)
    else:
      self.subset_type = None

    self.dataset = dataset
    self.feed = None
    self.pipeline_objects = []
    self.method_actions = []
    self.augmentations = []

    self.value_preprocessor = None
    self.padding_offset = None
    self.padding_target = None

    self.input_shape = self.dataset.sample_shape
    self.sample_count_to_shuffle = None
    if self.subset_type.is_training_set:
      self.sample_count_to_shuffle = self.dataset.ts_sample_count
    elif self.subset_type.is_validation_set:
      self.sample_count_to_shuffle = self.dataset.vs_sample_count
    elif self.subset_type.is_unknown_test_set:
      self.sample_count_to_shuffle = self.dataset.ut_sample_count
    self.batch_size  = None

    self._has_mapped_preprocessing_method = False
    self._is_multiclass = False

    self.feed = self.build_iterator()
    self.pipeline_objects.append(self.feed)
  # --------------------------------------------------------------------------------------------------------------------
  def multiclass(self):
    self._is_multiclass = True
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def normalize(self):
    self.value_preprocessor = Normalizer(self.dataset.name, self.dataset.filestore)
    if self.value_preprocessor.min is None:
      self.value_preprocessor.fit(self.dataset.ts_samples)
    self.method_actions.append("normalize")
    if not self._has_mapped_preprocessing_method:
      self.feed = self.build_preprocessor(self.feed)
      self.pipeline_objects.append(self.feed)
      self._has_mapped_preprocessing_method = True
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def map_preprocessing(self):
    if not self._has_mapped_preprocessing_method:
      self.feed = self.build_preprocessor(self.feed)
      self.pipeline_objects.append(self.feed)
      self._has_mapped_preprocessing_method = True
  # --------------------------------------------------------------------------------------------------------------------
  def standardize(self, axis_for_stats=None):
    self.value_preprocessor = Standardizer(self.dataset.name, self.dataset.filestore)
    if self.value_preprocessor.mean is None:
      self.value_preprocessor.fit(self.dataset.ts_samples, axis_for_stats=axis_for_stats)
    self.method_actions.append("standardize")
    self.map_preprocessing()
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def random_shuffle(self):
    self.feed = self.build_random_shuffler(self.feed)
    self.pipeline_objects.append(self.feed)
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def batch(self, batch_size):
    self.batch_size = batch_size
    self.feed = self.build_minibatch_maker(self.feed)
    self.pipeline_objects.append(self.feed)
    return self
# --------------------------------------------------------------------------------------------------------------------
  def augment_crop(self, padding_offset):
    self.padding_offset = padding_offset
    assert self.dataset.sample_shape is not None, "You should define the images input shape on the dataset"
    self.padding_target = self.dataset.sample_shape[0] + self.padding_offset
    self.map_preprocessing()
    self.feed = self.build_augmentation(self.feed, "random_crop")
    self.pipeline_objects.append(self.feed)
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def augment_flip_left_right(self):
    self.map_preprocessing()
    self.feed = self.build_augmentation(self.feed, "random_flip_left_right")
    self.pipeline_objects.append(self.feed)
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def augment_cutout(self):
    self.map_preprocessing()
    self.feed = self.build_augmentation(self.feed, "random_cutout")
    self.pipeline_objects.append(self.feed)
    return self
  # --------------------------------------------------------------------------------------------------------------------

  #// To be overrided \\
  # --------------------------------------------------------------------------------------------------------------------
  def build_iterator(self):
    return None
  # --------------------------------------------------------------------------------------------------------------------
  def build_preprocessor(self, feed):
    return feed
  # --------------------------------------------------------------------------------------------------------------------
  def add_augmentation(self, augmentation_kind):
    self.method_actions.add(augmentation_kind)
  # --------------------------------------------------------------------------------------------------------------------
  def build_random_shuffler(self, feed):
    return feed
# --------------------------------------------------------------------------------------------------------------------
  def build_minibatch_maker(self, feed):
    return feed
  # --------------------------------------------------------------------------------------------------------------------







# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2023-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
import os
import shutil
import sys
import re

from radnn.experiment import MLExperimentConfig
from radnn.experiment import get_experiment_code_ex, experiment_number_and_variation, experiment_code_and_timestamp
from radnn.core import now_iso
from radnn.system import FileSystem, FileStore
from radnn.system.tee_logger import TeeLogger








class MLExperimentEnv(dict):

  # --------------------------------------------------------------------------------------------------------------------
  @classmethod
  def experiment_filename_split(cls, filename):
    sTryFileName, sTryExt = os.path.splitext(filename)
    bIsVariationAndFold = "-" in sTryExt
    if bIsVariationAndFold:
      #LREXPLAINET22_MNIST_64.1-01
      sMainParts = filename.split("-")
      assert len(sMainParts) == 2, "Wrong experiment filename"
      sFoldNumber, _ = os.path.splitext(sMainParts[1])
      sParts = sMainParts[0].split("_")
      sModelName = f"{sParts[0]}_{sParts[1]}"
      sModelVar = sParts[2]
    else:
      sFileNameOnly, _ = os.path.splitext(filename)
      sMainParts = sFileNameOnly.split("-")
      if len(sMainParts) > 1:
        sFoldNumber = sMainParts[1]
      else:
        sFoldNumber = None
      sParts = sMainParts[0].split("_")
      sModelName = f"{sParts[0]}_{sParts[1]}"
      sModelVar = sParts[2]

    return sModelName, sModelVar, sFoldNumber
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  @classmethod
  def preload_config(cls, config_folder, experiment_group=None, experiment_base_name=None, experiment_variation=None, experiment_fold_number=None, experiment_filename=None):
    oPrintOutput = []
    oPrintOutput.append(f"[?] Experiment started at {now_iso()}")
    oPrintOutput.append(f" |__ {'model':<24}: {experiment_base_name}")

    dExperimentSpec = {"base_name": experiment_base_name, "variation": experiment_variation, "fold_number": experiment_fold_number}
    if experiment_filename is not None:
      # LREXPLAINET18_MNIST_08.1-01.json
      _, sFileNameFull = os.path.split(experiment_filename)
      experiment_base_name, experiment_variation, experiment_fold_number = cls.experiment_filename_split(sFileNameFull)
      experiment_fold_number = int(experiment_fold_number)
      dExperimentSpec = {"base_name": experiment_base_name, "variation": experiment_variation, "fold_number": experiment_fold_number}

    if experiment_group is not None:
      config_folder = os.path.join(config_folder, experiment_group)
    oConfigFS = FileStore(config_folder, must_exist=True)

    if "." in experiment_variation:
      sParts = experiment_variation.split(".")
      experiment_variation = f"{int(sParts[0]):02d}.{sParts[1]}"
    else:
      experiment_variation = f"{int(experiment_variation):02d}"
    sMessage = f" |__ {'variation':<24}: {experiment_variation}"

    if experiment_fold_number is not None:
      experiment_variation = f"{experiment_variation}-{experiment_fold_number:02d}"
      sMessage += f" fold: {experiment_fold_number}"
    oPrintOutput.append(sMessage)

    if experiment_filename is not None:
      sExperimentFileName = experiment_filename
    else:
      sExperimentFileName = oConfigFS.file(f"{experiment_base_name}_{experiment_variation}.json")
    oConfig = MLExperimentConfig(filename=sExperimentFileName, variation=experiment_variation)
    oPrintOutput.append(f" |__ {'configuration file':<24}: {sExperimentFileName}")

    return oConfig, oPrintOutput, dExperimentSpec
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, config_fs, model_fs=None, base_name=None, number=None, variation=None, fold_number=None,
               experiment_filename=None, experiment_code=None, experiment_config=None, model_filestore=None):

    super(MLExperimentEnv, self).__init__()
    if isinstance(config_fs, FileSystem):
      oConfigFS = config_fs.configs
      oModelFS  = config_fs.models
    else:
      oConfigFS = config_fs
      oModelFS = model_fs
      if oModelFS is None:
        oModelFS = config_fs

    # ...................... | Fields | ......................
    self.config_fs = oConfigFS
    self.model_fs  = oModelFS
    self.base_name = base_name
    self.number = number
    self.variation  = variation
    self.fold_number = fold_number

    if (experiment_filename is not None) or (experiment_code is not None):
      self.base_name, self.number, self.variation, self.fold_number, _ = self.determine_code_parts(experiment_filename,
                                                                      experiment_code, number, variation, fold_number)

    self.experiment_filename = experiment_filename
    if self.experiment_filename is None:
      sExperimentCode = get_experiment_code_ex(self.base_name, self.number, self.variation, self.fold_number)
      if experiment_code is not None:
        assert sExperimentCode == experiment_code, "Re-created experiment code mismatch."
      self.experiment_filename = self.config_fs.file(f"{sExperimentCode}.json")


    self.is_debugable = False
    self.is_retraining = False

    self._config = experiment_config
    if self._config is None:
      self._config = MLExperimentConfig(self.experiment_filename, number=self.number)
    self.experiment_fs = self.model_fs.subfs(self.experiment_code)
    # ........................................................

  # --------------------------------------------------------------------------------------------------------------------
  def save_config(self):
    self._config.save(self.experiment_filename)
  # --------------------------------------------------------------------------------------------------------------------
  def determine_code_parts(self, experiment_filename, experiment_code, number, variation, fold_number):
    sExperimentCode = None
    if experiment_code is not None:
      sExperimentCode = experiment_code
    elif experiment_filename is not None:
      _, sFileNameFull = os.path.split(experiment_filename)
      sExperimentCode, _ = os.path.splitext(sFileNameFull)

    sISODate = None
    if sExperimentCode is not None:
      oMatch = re.match(r"^\d{4}-\d{2}-\d{2}_\d{6}_", sExperimentCode)
      if oMatch is not None:
        sISODate = oMatch.group()[:-1]
        sExperimentCode = sExperimentCode[18:]

      base_name, variation, sFoldNumber = MLExperimentEnv.experiment_filename_split(sExperimentCode)
      if sFoldNumber is not None:
        fold_number = int(sFoldNumber)

    if number is None:
      number, variation = experiment_number_and_variation(variation)

    return base_name, number, variation, fold_number, sISODate
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def config(self):
    return self._config
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def experiment_code(self):
    return get_experiment_code_ex(self.base_name, self.number, self.variation, self.fold_number)
  # --------------------------------------------------------------------------------------------------------------------
  def copy_config(self, start_timestamp):
    sOriginalFileName = self.experiment_filename
    sNewFileName = self.experiment_fs.file(f"{start_timestamp}_{self.experiment_code}.json")
    shutil.copy(sOriginalFileName, sNewFileName)
  # --------------------------------------------------------------------------------------------------------------------
  def move_log(self, start_timestamp, original_filename):
    self.copy_config(start_timestamp)
    _, original_filename_only = os.path.split(original_filename)
    sNewFileName = f"{start_timestamp}_{self.experiment_code}.{original_filename_only}"
    shutil.move(original_filename, self.experiment_fs.file(sNewFileName))
    sys.stdout = TeeLogger(self.experiment_fs.file(sNewFileName))
  # --------------------------------------------------------------------------------------------------------------------
  '''
  def AssignSystemParams(self, p_oParamsDict):
    self.number = p_oParamsDict["ModelNumber"]
    self.is_debugable = p_oParamsDict["IsDebuggable"]
    self.is_retraining = p_oParamsDict["IsRetraining"]

    self._config = MLExperimentConfig(self.model_fs.file(self.experiment_code + ".json"), number=self.number)
  '''
  # --------------------------------------------------------------------------------------------------------------------
  def __str__(self)->str:
    sResult  = f"Experiment Code: {self.experiment_code} | Debugable: {self.is_debugable} | Retraining: {self.is_retraining}\n"
    sResult += f" |___ base name: {self.base_name} , number:{self.number} , variation:{self.variation}, fold_number: {self.fold_number}\n"
    sResult += f" |___ file name: {self.experiment_filename}\n"
    sResult += f"Models FileStore    : {self.model_fs}\n"
    sResult += f"Configs FileStore   : {self.config_fs}\n"
    sResult += f"Experiment FileStore: {self.experiment_fs}\n"
    sResult += f"Experiment Filename : {self.experiment_filename}\n"
    sResult += f"Configuration\n"
    sResult += "-"*40 + "\n"
    sResult += str(self._config) + "\n"
    return sResult
  # --------------------------------------------------------------------------------------------------------------------
  def __repr__(self)->str:
    return self.__str__()
  # --------------------------------------------------------------------------------------------------------------------





from radnn.learn.keras.keras_optimization_combo import KOptimizationCombo

class KLearningAlgorithm(object):
  # -----------------------------------------------------------------------------------
  def __init__(self, config, is_verbose=True):
    self.config = config
    self.is_verbose = is_verbose
    self._implementation = None

    self.prepare()
  # -----------------------------------------------------------------------------------
  @property
  def optimizer(self):
    oResult = None
    if self._implementation is not None:
      if isinstance(self._implementation, KOptimizationCombo):
        oResult = self._implementation.optimizer
    return oResult
  # -----------------------------------------------------------------------------------
  @property
  def callbacks(self):
    oResult = None
    if self._implementation is not None:
      if isinstance(self._implementation, KOptimizationCombo):
        oResult = self._implementation.callbacks
    return oResult
  # -----------------------------------------------------------------------------------
  def prepare(self):
    if mlsys.is_tensorflow_installed:
      self._implementation = KOptimizationCombo(self.config, self.is_verbose)
    return self
  # -----------------------------------------------------------------------------------



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import importlib
from PIL import Image
from .fileobject import FileObject


from PIL import Image
import numpy as np

class PNGFile(FileObject):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None):
    super(PNGFile, self).__init__(filename, parent_folder, error_template, "png")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None):
    filename = self._useFileName(filename)
    img = Image.open(filename).convert("RGB")
    return np.array(img).astype("uint8")
  # ----------------------------------------------------------------------------------
  def save(self, image_ndarray, filename=None):
    filename = self._useFileName(filename)

    if image_ndarray.dtype != np.uint8:
      image_ndarray = image_ndarray.astype(np.uint8)

    img = Image.fromarray(image_ndarray, mode="RGB")
    img.save(filename, format="PNG")
  # ----------------------------------------------------------------------------------

  


_is_opencv_installed = importlib.util.find_spec("cv2") is not None
if _is_opencv_installed:
  import cv2

  class PNGFile_CV2(FileObject):
    # ----------------------------------------------------------------------------------
    def __init__(self, filename, parent_folder=None, error_template=None):
      super(PNGFile_CV2, self).__init__(filename, parent_folder, error_template, "png")
    # ----------------------------------------------------------------------------------
    def load(self, filename=None):
      filename = self._useFileName(filename)
      nImage = cv2.imread(filename)
      nImageRGB = cv2.cvtColor(nImage, cv2.COLOR_BGR2RGB)
      return nImageRGB
    # ----------------------------------------------------------------------------------
    def save(self, image_ndarray, filename=None):
      filename = self._useFileName(filename)
      cv2.imwrite(filename, image_ndarray)
    # ----------------------------------------------------------------------------------


from .dataset_base import DataSetBase, DataSetCallbacks
from .dataset_base_legacy import CDataSetBase
from .sequence_dataset import SequenceDataSet
from .sample_set import SampleSet
from .sample_set_kind import SampleSetKind
from .sample_preprocessor import SamplePreprocessor

from .dataset_factory import DatasetFactory, DatasetBuildAdapter
from .constants import DataPreprocessingKind
from .custom_data_set import LegacyDataSet



from enum import Enum

class CNNSizeFactor(Enum):
  TINY = 1
  PETITE = 2
  SMALL = 3
  MEDIUM = 4


class CNNStemSetup(object):
  def __init__(self, size_factor: CNNSizeFactor, input_channels: int = 3, is_vit_like: bool = False):
    self.size_factor: CNNSizeFactor = size_factor
    self.input_channels = input_channels
    self.windows = []
    self.strides = []
    
    if is_vit_like:
      self.windows = [4]
      self.strides = [4]
    else:
      if self.size_factor is CNNSizeFactor.TINY:
        self.windows = [3]
        self.strides = [1]
      elif self.size_factor is CNNSizeFactor.PETITE:
        self.windows = [3]
        self.strides = [2]
      elif self.size_factor is CNNSizeFactor.SMALL:
        self.windows = [5]
        self.strides = [2]
      elif self.size_factor is CNNSizeFactor.MEDIUM:
        self.windows = [7, 3] #TODO: Check various papers
        self.strides = [2, 2]
      else:
        raise ValueError(f"Unknown CNN scale factor: {self.size_factor.name}")
    


from .plot_confusion_matrix import PlotConfusionMatrix
from .plot_learning_curve import PlotLearningCurve
from .plot_roc import PlotROC
from .plot_voronoi_2d import PlotVoronoi2D
from .plot_multi_scatter import MultiScatterPlot
from .plot_auto_multi_image import AutoMultiImagePlot
from .plot_function import PlotFunction
from .plot_histogram_of_classes import PlotHistogramOfClasses
from .plot_visualize_dataset2d import PlotVisualizeDataset2D
from .plot_legacy import CPlot, PlotDataset


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2019-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
import numpy as np
from sklearn import metrics
from radnn.utils import digit_count

# ==============================================================================================================================
class EvaluateClassification(object):
  # --------------------------------------------------------------------------------------------------------------
  def __init__(self, actual_classes, predicted_classes, probabilities=None):
    self.actual_classes = actual_classes
    self.predicted_classes = predicted_classes
    self.confusion_matrix = np.asarray(metrics.confusion_matrix(self.actual_classes, self.predicted_classes))

    self.accuracy = metrics.accuracy_score(self.actual_classes, self.predicted_classes)
    self.precision, self.recall, self.f1score, self.support = metrics.precision_recall_fscore_support(
      self.actual_classes, self.predicted_classes, average=None, zero_division=0.0)
    self.average_precision, self.average_recall, self.average_f1score, self.average_support = metrics.precision_recall_fscore_support(
      self.actual_classes, self.predicted_classes, average='weighted', zero_division=0.0)

    if probabilities is not None:
      self.auc = metrics.roc_auc_score(actual_classes, probabilities)
    else:
      self.auc = None


    self.true_neg = self.confusion_matrix[0][0]
    self.false_pos = self.confusion_matrix[0][1]
    self.false_neg = self.confusion_matrix[1][0]
    self.true_pos = self.confusion_matrix[1][1]
    
    self.class_indices, self.samples_per_class = np.unique(self.actual_classes, return_counts=True)
    self.class_count = None
  
  # --------------------------------------------------------------------------------------------------------------
  @property
  def confusion_matrix_normed(self):
    nNormed = self.confusion_matrix.copy().astype(np.float32)
    for nClassIndex, nSampleCount in enumerate(self.samples_per_class):
      if nSampleCount != 0:
        nNormed[nClassIndex, :] = nNormed[nClassIndex, :] / float(nSampleCount)
      else:
        nNormed[nClassIndex, :] = 0
    return nNormed
  # --------------------------------------------------------------------------------------------------------------
  def print_confusion_matrix(self, limit=40):
    nClassCount = len(self.confusion_matrix[0])
    sLabel = "Actual"
    sFiller = " "*len(sLabel)
    sIdent = " "*3
    nFormat = digit_count(nClassCount)
    sFormat = f"%" + str(nFormat) + "d"
    
    nLimitShrink = 2.0 / nFormat
    if nLimitShrink < 1.0:
      # Use the same horizontal space as for two digit class count.
      limit = int(limit * nLimitShrink)
    
    if nClassCount >= limit:
      nClassCount = limit
    print(f" {sIdent}{sFiller}Predicted  ")
    
    bAboveLimit = self.confusion_matrix.shape[0] > limit
    if bAboveLimit:
      sLine = "-" * (nFormat + 1) * limit
    else:
      sLine = "-" * (nFormat + 1) * nClassCount
    print(f" {sIdent}{sFiller}--{sLine}-")

    for nIndex, nRow in enumerate(self.confusion_matrix):
      if nIndex == limit:
        break
      else:
        sClassCorrect = " ".join([sFormat % x for x in nRow[:limit]])
        sClose = "|"
        if bAboveLimit:
          sClose = " ... "
        sTitle = sFiller
        if nIndex == 0:
          sTitle = sLabel
        print(f"{sIdent}{sTitle} | {sClassCorrect} {sClose}")
        
    if bAboveLimit:
      sMore = "." * nFormat + " "
      sLine = sMore * nClassCount
      print(f" {sIdent}{sFiller}  {sLine}")
    else:
      sLine = "-" * (nFormat + 1) * nClassCount
      print(f" {sIdent}{sFiller}--{sLine}-")
    print()
  # --------------------------------------------------------------------------------------------------------------
  def format_series_as_pc(self, metric_series):
    oValStr = []
    for x in metric_series:
      sX = f"{x*100.0:.2f}"
      oValStr.append(sX)

    oValues = [f"{x:^7}" for x in oValStr]
    return " |".join(oValues)
  # --------------------------------------------------------------------------------------------------------------
  def print_per_class(self, class_names=None):
    if class_names is not None:
      nClassCount = len(class_names.keys())
      oClasses = [f"{class_names[x]:7}" for x in list(range(nClassCount))]
    else:
      oClasses = sorted(np.unique(self.actual_classes))
      nClassCount = len(oClasses)
      oClasses = [f"{x:^7}" for x in oClasses]
    self.class_count = nClassCount

    sClasses = " |".join(oClasses)
    nRepeat = 28 + (7+2)*self.class_count
    print(f"                            |{sClasses}|")
    print("-"*nRepeat)
    print(f"Per Class Recall %          |{self.format_series_as_pc(self.recall[:])}|")
    print(f"Per Class Precision %       |{self.format_series_as_pc(self.precision[:])}|")
    print("-" * nRepeat)
  # --------------------------------------------------------------------------------------------------------------
  def print_overall(self):
    print(f"Accuracy %                  :{self.accuracy*100.0       :.3f}")
    print(f"Average F1 Score %          :{self.average_f1score*100.0:.3f}")
    print(f"Weighted Average Recall %   :{self.average_recall*100.0:.3f}")
    print(f"Weighted Average Precision %:{self.average_precision*100.0:.3f}")
    if (self.class_count == 2) and (self.auc is not None):
      print(f"Area Under the Curve (AUC):{self.auc:.4f}")
    print()
  # --------------------------------------------------------------------------------------------------------------


# ==============================================================================================================================


# alias for compatibility
class CEvaluator(EvaluateClassification):
  pass




import torch
import torch.nn as nn


class RMSELoss(nn.Module):
  def __init__(self, eps=1e-7):
    super().__init__()
    self.mse = nn.MSELoss()
    self.eps = eps
  
  def forward(self, yhat, y):
    loss = torch.sqrt(self.mse(yhat, y) + self.eps)
    return loss



from radnn.core import RequiredLibs
oReqs = RequiredLibs()
if oReqs.is_tensorflow_installed:
  import tensorflow.keras as ker

class KLearningRateScheduler(ker.callbacks.LearningRateScheduler):
  # -----------------------------------------------------------------------------------
  def __init__(self, config, is_verbose=True):
    self.config = None
    self.lr_schedule = None
    self.is_verbose = is_verbose

    if isinstance(config, dict):
      self.config = config
      self.lr_schedule = config["Training.LearningRateSchedule"]
    elif isinstance(config, list):
      self.lr_schedule = config

    super(KLearningRateScheduler, self).__init__(self.check_schedule)
  # -----------------------------------------------------------------------------------
  def check_schedule(self, epoch, lr):
    nNewLR = lr

    for nIndex, oSchedule in enumerate(self.lr_schedule):
      if epoch == oSchedule[0]:
        nNewLR = oSchedule[1]
        if self.is_verbose:
          print("Schedule #%d: Setting LR to %.5f" % (nIndex + 1, nNewLR))
        break

    if self.is_verbose:
      print("LR: %.6f" % nNewLR)
    return nNewLR
  # -----------------------------------------------------------------------------------


from .ml_experiment_config import get_experiment_code, get_experiment_code_ex, experiment_number_and_variation, experiment_code_and_timestamp
from .ml_experiment_config import MLExperimentConfig
from .ml_experiment_env import MLExperimentEnv
from .ml_experiment_log import MLExperimentLog
from .identification import experiment_fold_number, experiment_name_with_fold




import numpy as np

class DescriptiveStats(object):
  INNER_FENCE_RATIO = 1.5
  OUTER_FENCE_RATIO = 3.0

  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, data):
    if (data.ndim == 2) and (np.prod(data.shape) == np.max(data.shape)):
      data = data.reshape(-1)

    self.min = np.min(data)
    self.max = np.max(data)

    self.mean = np.mean(data)
    self.std = np.std(data)

    self.q1 = np.percentile(data, q=25)
    self.median = np.median(data)
    self.q2 = self.median
    self.q3 = np.percentile(data, q=75)
    self.iq_range = self.q3 - self.q1

    self.inner_fence_low = self.q1 - type(self).INNER_FENCE_RATIO * self.iq_range
    self.inner_fence_high = self.q3 + type(self).INNER_FENCE_RATIO * self.iq_range

    self.outer_fence_low = self.q1 - type(self).OUTER_FENCE_RATIO * self.iq_range
    self.outer_fence_high = self.q3 + type(self).OUTER_FENCE_RATIO * self.iq_range

  # --------------------------------------------------------------------------------------------------------------------
  def outliers(self, data, is_using_outer_fence=False, dtype=np.float32):
    if is_using_outer_fence:
      nResult = np.asarray([x for x in data if x < self.outer_fence_low or x > self.outer_fence_high], dtype=dtype)
    else:
      nResult = np.asarray([x for x in data if x < self.inner_fence_low or x > self.inner_fence_high], dtype=dtype)
    return nResult
  # --------------------------------------------------------------------------------------------------------------------
  def __str__(self):
    sResult = "range=[%.6f,%.6f] mean=%6f std=%.6f iqrange=%.6f Q1=%.6f median (Q2)=%.6f Q3=%.6f max (Q4)=%.6f" % (self.min, self.max,
                                                                                                self.mean, self.std,
                                                                                                self.iq_range,
                                                                                                self.q1, self.median,
                                                                                                self.q3, self.max)
    return sResult
  # --------------------------------------------------------------------------------------------------------------------


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2024-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import sys
from datetime import datetime


class TeeLogger(object):
  # --------------------------------------------------------------------------------------------------------
  def __init__(self, file_name):
    self.last_flush = datetime.now()
    sMode = "w"
    if os.path.exists(file_name):
      sMode = "a"
    self.file = open(file_name, sMode)
    self.stdout = sys.stdout
    self.isclosed = False
# --------------------------------------------------------------------------------------------------------
  def write(self, message):
    self.stdout.write(message)
    self.file.write(message)
    dNow = datetime.now()
    oDelta = dNow - self.last_flush
    if oDelta.total_seconds() >= 5:
      self.flush()
      self.last_flush = dNow
  # --------------------------------------------------------------------------------------------------------
  def flush(self):
    if not self.isclosed:
      self.stdout.flush()
      self.file.flush()
  # --------------------------------------------------------------------------------------------------------
  def close(self):
    self.isclosed = True
    self.stdout.close()
    self.file.close()
  # --------------------------------------------------------------------------------------------------------


# ......................................................................................
# MIT License

# Copyright (c) 2019-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
class SubsetType(object):
  def __init__(self, name):
    self.name = name.lower()
    self.type = -1 # Unknown
    if self.is_training_set:
      self.type = 0
    elif self.is_validation_set:
      self.type = 1
    elif self.is_unknown_test_set:
      self.type = 2
  @property
  def is_training_set(self):
    return (self.name == "training") or (self.name == "train") or (self.name == "ts")

  @property
  def is_validation_set(self):
    return (self.name == "validation") or (self.name == "val") or (self.name == "vs")

  @property
  def is_unknown_test_set(self):
    return (self.name == "testing") or (self.name == "test") or (self.name == "ut")


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
from .semaphore_lock import SemaphoreLock

    

class ThreadSafeQueue(object):
    #-------------------------------------------------------------------------------------------------------------------
    def __init__(self, name="Queue", max_queued_items=None):
        #........................... |  Instance Attributes | ...........................
        self.name = name
        self.queue=[]
                
        # // Settings \\
        self.max_queued_items = max_queued_items
        self.can_delete              = True
        
        # // Control Params \\
        self.is_queue_locked = False
        
        # // Composites \\
        self.update_lock = SemaphoreLock()
        #................................................................................

    # -------------------------------------------------------------------------------------------------------------------
    @property
    def count(self):
        return len(self.queue)

    # -------------------------------------------------------------------------------------------------------------------
    @property
    def is_empty(self):
        bResult = (len(self.queue) == 0)

        return bResult
    #-------------------------------------------------------------------------------------------------------------------
    def push(self, value):
        self.update_lock.lock()
        try:
            if self.max_queued_items is None:
                bCanAppend = True
            else: 
                bCanAppend = len(self.queue) < self.max_queued_items
            
            if bCanAppend:     
                self.queue.append(value)
        finally:
            self.update_lock.unlock()
    #-------------------------------------------------------------------------------------------------------------------
    def pop(self, index=0):
        nValue = None        
        if self.can_delete:
            self.update_lock.lock()
            try:
                bMustPop = len(self.queue) > 0
                if bMustPop:
                    if index == -1:
                        index = len(self.queue) - 1
                    nValue = self.queue.pop(index)
            finally:
                self.update_lock.unlock()
            
        return nValue
    #-------------------------------------------------------------------------------------------------------------------
    def clear(self):
        self.update_lock.lock()
        try:
            self.queue = []
        finally:
            self.update_lock.unlock()
    #-------------------------------------------------------------------------------------------------------------------
    '''
    def pop_ex(self, index=0):
        nValue,nRemainingItemCount = [None,None]
        if self.can_delete:
            self.update_lock.Lock()
            try:
                nRemainingItemCount = len(self.queue)
                bMustPop = nRemainingItemCount > 0 
                if bMustPop:
                    if index == -1:
                        index = len(self.queue) - 1
                    nValue = self.queue.pop(index)
                    nRemainingItemCount = len(self.queue)
            finally:
                self.update_lock.UnLock()
            
        return nValue, nRemainingItemCount
    '''
    #-------------------------------------------------------------------------------------------------------------------
    def push_ex(self, message):
        while self.is_queue_locked:
            pass
        
        if self.max_queued_items is None:
            bCanAppend = True
        else: 
            bCanAppend = len(self.queue) < self.max_queued_items
            
        if not bCanAppend:
            return  
        
        self.is_queue_locked = True
        try:
            self.queue.append(message)
        finally:
            self.is_queue_locked = False
    #-------------------------------------------------------------------------------------------------------------------
    def pop_ex(self):
        sMessage = None
        if self.can_delete:
            while self.is_queue_locked:
                pass
                
            self.is_queue_locked = True
            try:
                if len(self.queue) > 0:
                    sMessage = self.queue.pop(0)
            finally:
                self.is_queue_locked = False

        return sMessage
    #-------------------------------------------------------------------------------------------------------------------
















# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import glob

class FileObject(object):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None, default_file_extension=None):
    self.filename = filename
    self.parent_folder = parent_folder
    self.error_template = error_template
    self.default_file_extension = default_file_extension
    if self.default_file_extension is not None:
      if not self.default_file_extension.startswith("."):
        self.default_file_extension = "." + self.default_file_extension
    self.filename       = filename
    self.parent_folder  = parent_folder
    self.error_template = error_template
    self.default_file_extension = default_file_extension
    if self.default_file_extension is not None:
      if not self.default_file_extension.startswith("."):
        self.default_file_extension = "." + self.default_file_extension
  # ----------------------------------------------------------------------------------
  def close(self):
    pass
  # ----------------------------------------------------------------------------------
  def __enter__(self):
    return self
  # ----------------------------------------------------------------------------------
  def __exit__(self, exc_type, exc_value, traceback):
    self.close()
    return True
  # ----------------------------------------------------------------------------------
  def _useFileName(self, filename):
    if filename is not None:
        self.filename = filename
    if self.parent_folder is not None:
      sFilename = os.path.join(self.parent_folder, self.filename)
    else:
      sFilename = self.filename
    return sFilename
  # ----------------------------------------------------------------------------------
  def list_files(self, file_matching_pattern=None, is_full_path=True, is_removing_extension=False, sort_filename_key=None):
    if file_matching_pattern is None:
      file_matching_pattern = self.default_file_extension

    if file_matching_pattern is None:
      file_matching_pattern = "*.*"
    elif not file_matching_pattern.startswith("*."):
      file_matching_pattern = "*." + file_matching_pattern
    elif file_matching_pattern.startswith("."):
      file_matching_pattern = "*" + file_matching_pattern

    sEntries = glob.glob1(self.parent_folder, file_matching_pattern)

    if is_removing_extension:
      oFileNamesOnly = []
      for sEntry in sEntries:
        sFileNameOnly, _ = os.path.splitext(sEntry)
        oFileNamesOnly.append(sFileNameOnly)
      sEntries = sorted(oFileNamesOnly, key=sort_filename_key)

    if is_full_path:
      oResult = [os.path.join(self.parent_folder, x) for x in sEntries]
    else:
      oResult = [x for x in sEntries]

    return oResult
  # --------------------------------------------------------------------------------------------------------


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2022-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

class MultiScatterPlot(object):
  # --------------------------------------------------------------------------------------
  # Constructor
  def __init__(self, title, data=None, class_count=10, class_names=None):
    # ................................................................
    # // Fields \\
    self.Title = title
    self.Data = data

    self.ClassCount = class_count
    if class_names is not None:
      self.ClassNames = class_names
      self.ClassCount = len(self.ClassNames)

    if (self.ClassCount <= 10):
      self.ColorMap = cm.get_cmap("tab10")
    elif (self.ClassCount <= 20):
      self.ColorMap = cm.get_cmap("tab20")
    else:
      self.ColorMap = cm.get_cmap("prism")
      # self.ColorMap = colors.ListedColormap(["darkorange","darkseagreen"])

    self.PlotDimensions = [14, 8]
    self.PointSize = 48

    self.PanesPerRow = 2
    if self.Data is not None:
      self.Panes = len(self.Data)

    self.__fig = None
    self.__ax = None
    # ................................................................

  # --------------------------------------------------------------------------------------
  def add_data(self, dataset_name, samples, labels=None):
    if self.Data is None:
      self.Data = []

    if labels is None:
      labels = np.zeros((samples.shape[0]), np.int32)
    self.Data.append([dataset_name, samples, labels])
    self.Panes = len(self.Data)

  # --------------------------------------------------------------------------------------
  def prepare(self, index, x_axis_caption, y_axis_caption):
    # The 2D dataset for the scatter plot
    sDataName, nSamples, nLabels = self.Data[index]
    nXValues = nSamples[:, 0]
    nYValues = nSamples[:, 1]
    nLabels = nLabels

    if self.__fig is None:
      if self.Panes == 1:
        self.__fig, self.__ax = plt.subplots(nrows=1, ncols=1, sharex=False, sharey=False, squeeze=False,
                                             figsize=self.PlotDimensions)
      else:
        nRows = self.Panes // self.PanesPerRow
        if (self.Panes % self.PanesPerRow != 0):
          nRows += 1
        self.__fig, self.__ax = plt.subplots(nrows=nRows, ncols=self.PanesPerRow, sharex=False, sharey=False,
                                             squeeze=False, figsize=self.PlotDimensions)

    nRow = index // self.PanesPerRow
    nCol = index - (nRow * self.PanesPerRow)
    oPlot = self.__ax[nCol, nRow]
    oPlot.set_xlabel(x_axis_caption)
    oPlot.set_ylabel(y_axis_caption)
    oPlot.set_title(sDataName)

    oScatter = oPlot.scatter(nXValues, nYValues, s=self.PointSize, c=nLabels, cmap=self.ColorMap)

    oLegend = oPlot.legend(*oScatter.legend_elements(), loc="lower right", title="Classes", framealpha=0.4,
                           labelspacing=0.1)
    oPlot.add_artist(oLegend)

    if index == self.Panes - 1:
      plt.title(self.Title)
      plt.tight_layout(pad=1.01)

    return self
  # --------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2022-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors

class PlotVisualizeDataset2D(object):
  # --------------------------------------------------------------------------------------------------------------------
  # Constructor
  def __init__(self, samples_2d, labels, title=None):
    # ................................................................
    # // Fields \\
    self.title = title
    self.samples = samples_2d
    self.labels = labels
    # ................................................................
  # --------------------------------------------------------------------------------------------------------------------
  def prepare(self, title=None, figure_size=(8, 8),
              is_min_max_scaled=False, line_slope=None, line_intercept=None,
              limits_x=[-4, 4], limits_y=[-4, 4]):
    if title is not None:
      self.title = title
      
    # Two dimensional dataset for the scatter plot
    nXValues = self.samples[:, 0]
    nYValues = self.samples[:, 1]
    nLabels = self.labels
    
    # https://matplotlib.org/3.1.0/gallery/color/named_colors.html
    oColors = ["darkorange", "darkseagreen"]
    oLabelDescriptions = ["orange tree", "olive tree"]
    oColorMap = colors.ListedColormap(oColors)
    
    fig, ax = plt.subplots(figsize=figure_size)
    plt.scatter(nXValues, nYValues, c=nLabels, cmap=oColorMap)
    
    plt.title(self.title)
    cb = plt.colorbar()
    nLoc = np.arange(0, max(nLabels), max(nLabels) / float(len(oColors)))
    cb.set_ticks(nLoc)
    cb.set_ticklabels(oLabelDescriptions)
    
    if (line_slope is not None):
      x1 = np.min(nXValues)
      y1 = line_slope * x1 + line_intercept;
      x2 = np.max(nXValues)
      y2 = line_slope * x2 + line_intercept;
      oPlot1 = ax.plot([x1, x2], [y1, y2], 'r--', label="Decision line")
      oLegend = plt.legend(loc="upper left", shadow=True, fontsize='x-large')
      oLegend.get_frame().set_facecolor("lightyellow")
    
    if is_min_max_scaled:
      ax.set_xlim((-0.05, 1.05))
      ax.set_ylim((-0.05, 1.05))
    else:
      ax.set_xlim(limits_x[0], limits_x[1])
      ax.set_ylim(limits_y[0], limits_y[1])
    
    ax.set_xlabel('Feature 1')
    ax.set_ylabel('Feature 2')
    
    return self
  
  # --------------------------------------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------------------------------------




from radnn import FileStore
class MLExperimentStore(FileStore):
  def __init__(self, base_folder, is_verbose=False, must_exist=False):
    super(MLExperimentStore, self).__init__(base_folder, is_verbose, must_exist)

    self.param_fs = self.subfs("weights")
    self.log_fs = self.subfs("logs")
    self.checkpoint_fs = self.subfs("checkpoints")
    self.state_fs = self.subfs("state")



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2023-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
from sklearn import metrics
import matplotlib.pyplot as plt

'''
    Receiver Operator Characteristics (ROC) Plot
'''
class PlotROC(object):
  # --------------------------------------------------------------------------------------
  def __init__(self, actual_labels, predicted_probs, title="Receiving Operator Characteristics Curve"):
    self.actual_labels = actual_labels
    self.predicted_probs = predicted_probs
    self.title = title
  # --------------------------------------------------------------------------------------
  def prepare(self, true_threshold=0.5, figure_size=[6.00, 6.00], is_showing_grid=True):
    plt.clf()
    plt.rcParams["figure.figsize"] = figure_size
    plt.rcParams["figure.autolayout"] = True

    nFPR, nTPR, nThresholds = metrics.roc_curve(self.actual_labels, self.predicted_probs)
    nAUC = metrics.roc_auc_score(self.actual_labels, self.predicted_probs)
    plt.xlim(0, 1.02)
    plt.ylim(0, 1.02)
    if is_showing_grid:
      plt.grid()
    plt.plot(nFPR, nTPR, label="ROC curve (AUC=%.2f)" % nAUC, linewidth=2)
    plt.plot([0.0, 1.0], [0.0, 1.0], 'r--', label="Random prediction")  # , color="yellow", linewidth=1)

    plt.ylabel("TPR (True Positive Rate)")
    plt.xlabel("FPR (False Positive Rate)")

    nPointIndex = None
    for nThresholdIndex, nThreshold in enumerate(nThresholds):
      if nThreshold < true_threshold:
        nPointIndex = nThresholdIndex
        break

    if nPointIndex is not None:
      plt.plot(nFPR[nPointIndex], nTPR[nPointIndex]
              , label="Threshold %.2f" % nThreshold
              , marker="o", markersize=10, markeredgecolor="blue", markerfacecolor="yellow")
      plt.title(self.title, fontsize=18)
      plt.legend(loc=4)
    return self
  # --------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------
